/*
 *	flashsign.c
 *
 * Copyright (c) 2001, 2007
 * 	Koh-ichi Ito All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the developer nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE DEVELOPPER AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Label.h>
#include <X11/Xlib.h>
#include <X11/Xlocale.h>
#include <arpa/inet.h>
#include <ctype.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include "queue.h"
#define INTERVAL	250
#define MAXLABELLEN	256	/* shown in labelWidget */
#define MAXTEXTLEN	2048	/* scrolled text */
#define WIDTH		65
#define GAP		5	/* gap between texts when wrapping */
#define max(x, y)	( (x) < (y)? (y):(x) )

void	wakeup(),
	lshift(),
	quit(),
	errquit(),
	cleanup(),
	usage();
int	storepidfile();
long	Interval;	/* Interval timer in XtAppMainLoop() */
int	Fifo,
	Sock,
	Width,		/* width in characters, not in dots */
	RemainFifo,
	RaiseOnNew;
char	*MyName,
	*FifoName,
	*PidFile;
struct sockaddr_in	SockAddr;
static char	Id[] = "$Id: flashsign.c,v 1.2 2007/08/16 02:32:45 kohi Exp $";

main(argc, argv)

	int	argc;
	char	**argv;
{
	XtAppContext	appcon;
	Widget	toplevel,
		label;
	int	i;
	ushort	udp_port;
	char	buf[MAXPATHLEN],	/* a scratch buffer */
		spaceholder[MAXLABELLEN];	/* for label Widget */
	struct stat	dummy;
	static String resource[] = {
		"*label.fontSet: -*--24-*",
		"*label.International: True",
		NULL
	};
	static XrmOptionDescRec	options[] = {
		{"-fn", "*label.fontSet", XrmoptionSepArg, NULL},
	};

	Interval = INTERVAL;
	Width = WIDTH;
	RemainFifo = 0;
	FifoName = NULL;
	PidFile = NULL;
	udp_port = 0;
	RaiseOnNew = 0;
	if ( (MyName = strrchr(argv[0], '/') ) ) {
		MyName++;
	} else {
		MyName = argv[0];
	}
	strncpy(buf, MyName, MAXPATHLEN);
	if (islower(buf[0] ) ) {
		buf[0] = toupper(buf[0] );
	}
	XtSetLanguageProc(NULL, NULL, NULL);
	toplevel = XtAppInitialize(&appcon, buf,
		options, XtNumber(options),
		&argc, argv, resource, NULL, 0);
	argc--;
	argv++;
	while ( (0 < argc) && (**argv == '-') ) {
		switch (*++*argv) {
		case 's':
			if (argc <= 1) {
				usage();
			}
			argc--;
			argv++;
			Interval = atol(*argv);
			break;
		case 'F':
			RemainFifo = 1;
		case 'f':
			if (argc <= 1) {
				usage();
			}
			argc--;
			argv++;
			FifoName = *argv;
			break;
		case 'u':
			if (argc <= 1) {
				usage();
			}
			argc--;
			argv++;
			udp_port = atoi(*argv);
			break;
		case 'w':
			if (argc <= 1) {
				usage();
			}
			argc--;
			argv++;
			Width = atoi(*argv);
			if (MAXLABELLEN <= Width) {
				fprintf(stderr, "%s: width too much.\n",
					MyName);
				exit(1);
			}
			break;
		case 'p':
			if (argc <= 1) {
				usage();
			}
			argc--;
			argv++;
			PidFile = *argv;
			if (!storepidfile() ) {
				PidFile = NULL;
			}
			break;
		case 'r':
			RaiseOnNew = 1;
			break;
		default:
			usage();
		}
		argc--;
		argv++;
	}
	if (FifoName) {
		if (stat(FifoName, &dummy) < 0) {
			if (mkfifo(FifoName, 0600) < 0) {
				snprintf(buf, 1024, "%s: mkfifo(%s)",
					 MyName, FifoName);
				perror(buf);
				exit(1);
			}
		}
		if ( (Fifo = open(FifoName, O_RDONLY | O_NONBLOCK) ) < 0) {
			snprintf(buf, 1024, "%s: open(%s)",
				 MyName, FifoName);
			perror(buf);
			exit(1);
		}
	}
	if (udp_port) {
		if ( (Sock = socket(PF_INET, SOCK_DGRAM, 0) ) < 0) {
			snprintf(buf, 1024, "%s: socket(%s)",
				 MyName, FifoName);
			perror(buf);
			if (Fifo) {
				close(Fifo);
			}
			exit(1);
		}
		bzero(&SockAddr, sizeof(SockAddr) );
		SockAddr.sin_family = AF_INET;
		SockAddr.sin_addr.s_addr = htonl(INADDR_ANY);
		SockAddr.sin_port = htons(udp_port);
		if (bind(Sock, (struct sockaddr *)&SockAddr,
			sizeof(SockAddr) ) < 0) {
			
			snprintf(buf, 1024, "%s: bind(%s)",
				 MyName, FifoName);
			perror(buf);
			close(Sock);
			if (Fifo) {
				close(Fifo);
			}
			exit(1);
		}
	}
	if (!Fifo && !udp_port) {
		fprintf(stderr,
			"%s: neigher FIFO nor UDP port is specified.\n",
			MyName);
		exit(1);
	}
	for (i = 0; i < Width - 1; i++) {
		spaceholder[i] = ' ';
	}
	spaceholder[Width - 1] = '\0';
	if (0 < argc) {
		enqueue(*argv);
		argc--;
		argv++;
	} else {
		enqueue("");
	}
	label = XtVaCreateManagedWidget("label", labelWidgetClass,
		toplevel, XtNlabel, spaceholder, NULL);
	signal(SIGINT, quit);
	signal(SIGTERM, quit);
	/* signal(SIGWINCH, wish_to_do_something);*/
	XtAppAddTimeOut(XtWidgetToApplicationContext(label),
		Interval, wakeup, (XtPointer)label);
	XtRealizeWidget(toplevel);
	XtAppMainLoop(appcon);
}

void wakeup(labelWidget, id)

	XtPointer	labelWidget;
	XtIntervalId	*id;
{
	struct timeval	timeout;
	fd_set	ifd;
	static char	labelstr[MAXLABELLEN] = "",
			*curtext = NULL;
	char	buf[MAXTEXTLEN],	/* a scratch buffer */
		*p;
	static int	k = 0,	/* sweep CurText from left to right */
			len = 0;
	int	i,
		n,
		in_kanji;
	socklen_t	fromlen;

	timeout.tv_sec = 0L;
	timeout.tv_usec = 0L;
	fromlen = sizeof(SockAddr);
	FD_ZERO(&ifd);
	if (Fifo) {
		FD_SET(Fifo, &ifd);
	}
	if (Sock) {
		FD_SET(Sock, &ifd);
	}
	if (select(max(Fifo, Sock) + 1, &ifd, NULL, NULL, &timeout) ) {
		if ( (FD_ISSET(Fifo, &ifd)
		   && (0 < (n = read(Fifo, buf, MAXTEXTLEN) ) ) )
		  || (FD_ISSET(Sock, &ifd)
		   && (0 < (n = recvfrom(Sock, buf, MAXTEXTLEN,
				0, (struct sockaddr *)&SockAddr,
				&fromlen) ) ) ) ) {
			buf[n] = '\0';
			for (p = buf; *p; p++) {
				if ( (*p == '\r') || (*p == '\n') ) {
					*p = ' ';
				}
			}
			if (!enqueue(buf) ) {
				snprintf(buf, MAXTEXTLEN,
					 "%s: enqueue()", MyName);
				perror(buf);
				free(curtext);
				errquit();
			}
		}
	}
	if (!curtext) {
		if (!(curtext = dequeue() ) ) {
			fprintf(stderr,
				"%s: couldn't get initial text from queue.\n",
				MyName);
			errquit();
		}
		len = strlen(curtext);
		for (i = 0; i < Width - 1; i++) {
			labelstr[i] = ' ';
		}
		labelstr[Width - 1] = '\0';
	}
	if (len + GAP <= k) {
		k = 0;
		if (topqueue() ) {
			free(curtext);
			curtext = dequeue();
			len = strlen(curtext);
			if (RaiseOnNew) {
				XRaiseWindow(XtDisplay(labelWidget),
					XtWindow(XtParent(labelWidget) ) );
			}
		}
	}
	in_kanji = labelstr[0] & 0x80;		/* dirty, EUC kanji depend */
	if (k < len) {
		lshift(labelstr, curtext[k++] );
	} else if (k < len + GAP) {
		lshift(labelstr, ' ');
		k++;
	}
	if (in_kanji) {
		labelstr[0] = ' ';
	}
	labelstr[Width - 1] = '\0';
	XtVaSetValues(labelWidget, XtNlabel, labelstr, NULL);
	XFlush(XtDisplay(labelWidget) );
	XtAppAddTimeOut(XtWidgetToApplicationContext(labelWidget),
		Interval, wakeup, labelWidget);
}

void lshift(s, c)

	char	*s,
		c;
{
	char	*p;

	p = s + 1;
	while (*p) {
		/*
		 * I experienced the problem by the code
		 *
		 *	*(p - 1) = *p++;
		 *
		 * with gcc version 3.4.6 [FreeBSD] 20060305
		 */
		*(p - 1) = *p;
		p++;
	}
	*(p - 1) = c;
}

int storepidfile()
{
	FILE	*fp;

	if (!(fp = fopen(PidFile, "w") ) ) {
		return 0;
	}
	fprintf(fp, "%d", getpid() );
	fclose(fp);
	return 1;
}

void cleanup()
{
	if (Fifo) {
		close(Fifo);
	}
	if (FifoName && !RemainFifo) {
		unlink(FifoName);
	}
	if (PidFile) {
		unlink(PidFile);
	}
	if (Sock) {
		close(Sock);
	}
}

void quit()
{
	cleanup();
	exit(0);
}

void errquit()
{
	cleanup();
	exit(1);
}

void usage()
{
	fprintf(stderr,
		"%s: [-s speed][-w width][-F FIFO|-f fifo|-u port][-r][text]\n",
		MyName);
	exit(1);
}
